(* This is the source calculus of the sample client. It is a core ML. *)

(* Nominal representation of term variables, type variables and binders. *)

type 'a list = 'a List.t =
  | []
  | (::) of 'a * 'a list
  [@@deriving compare]

type range =
  Lexing.position * Lexing.position

let compare_range _ _ = 0

type tevar = String.t [@@deriving compare]
type tyvar = String.t [@@deriving compare]

type typ =
  | TyVar of range * tyvar
  | TyArrow of range * typ * typ
  | TyProduct of range * typ list
  | TyConstr of range * Datatype.TyConstrId.t * typ list
 [@@deriving compare]

type type_annotation = tyvar list * typ [@@deriving compare]
  (* some <flexible vars> . <ty> *)

type term =
  | Var of range * tevar
  | Hole of range * term list
  | Abs of range * tevar * term
  | App of range * term * term
  | Let of range * tevar * term * term
  | Annot of range * term * type_annotation
  | Tuple of range * term list
  | LetProd of range * tevar list * term * term
  | Variant of range * Datatype.LabelId.t * term Option.t
  | Match of range * term * branch list

and branch = pattern * term

and pattern =
  | PVar of range * tevar
  | PWildcard of range
  | PAnnot of range * pattern * type_annotation
  | PVariant of range * Datatype.LabelId.t * pattern Option.t
  | PTuple of range * pattern list
  [@@deriving compare]

type datatype_env = (tyvar, typ option) Datatype.Env.t

let dummy_pos = (Lexing.( dummy_pos, dummy_pos ))
